import abc
import threading

from idempotency_key import utils


class IdempotencyKeyLock(abc.ABC):
    @abc.abstractmethod
    def acquire(self, *args, **kwargs) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def release(self):
        raise NotImplementedError()


class ThreadLock(IdempotencyKeyLock):
    """
    Should be used only when there is one process sharing the storage class resource.
    This uses the built-in python threading module to protect a resource.
    """

    storage_lock = threading.Lock()

    def acquire(self, *args, **kwargs) -> bool:
        return self.storage_lock.acquire(
            blocking=True, timeout=utils.get_lock_timeout()
        )

    def release(self):
        self.storage_lock.release()



import abc
from collections import defaultdict
import pickle
from typing import Tuple

from django.core.cache import caches


class IdempotencyKeyStorage(object):
    @abc.abstractmethod
    def store_data(self, cache_name: str, encoded_key: str, response: object) -> None:
        """
        called when data should be stored in the storage medium
        :param cache_name: The name of the cache to use defined in settings under CACHES
        :param encoded_key: the key used to store the response data under
        :param response: The response data to store
        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_data(self, cache_name: str, encoded_key: str) -> Tuple[bool, object]:
        """
        Retrieve data from the sore using the specified key.
        :param cache_name: The name of the cache to use defined in settings under CACHES
        :param encoded_key: The key that was used to store the response data
        :return: the response data
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def validate_storage(name: str):
        """
        Validate that the storage name exists. If the class is using django `CACHES` setting then this function ensures
        that the cache is setup correctly in the settings file and will cause a failure at startup if it is not.
        This function should raise an exception if the storage name cannot be validated.
        :param name: The name of the storage.
        """
        raise NotImplementedError


class MemoryKeyStorage(IdempotencyKeyStorage):
    def __init__(self):
        self.idempotency_key_cache_data = defaultdict(dict)

    def store_data(self, cache_name: str, encoded_key: str, response: object) -> None:
        self.idempotency_key_cache_data[cache_name][encoded_key] = response

    def retrieve_data(self, cache_name: str, encoded_key: str) -> Tuple[bool, object]:
        the_cache = self.idempotency_key_cache_data.get(cache_name)
        if the_cache and encoded_key in the_cache.keys():
            return True, the_cache[encoded_key]

        return False, None

    @staticmethod
    def validate_storage(name: str):
        pass


class CacheKeyStorage(IdempotencyKeyStorage):
    def store_data(self, cache_name: str, encoded_key: str, response: object) -> None:
        str_response = pickle.dumps(response)
        caches[cache_name].set(encoded_key, str_response)

    def retrieve_data(self, cache_name: str, encoded_key: str) -> Tuple[bool, object]:
        if encoded_key in caches[cache_name]:
            str_response = caches[cache_name].get(encoded_key)
            return True, pickle.loads(str_response)

        return False, None

    @staticmethod
    def validate_storage(name: str):
        # Check that the cache exists. If the cache is not found then an InvalidCacheBackendError is raised.
        # Not that there is no get function on the caches object so we cannot perform a normal check.
        caches[name]



from django.conf import settings
from django.utils import module_loading

from idempotency_key import status


def idempotency_key_exists(request):
    return getattr(request, "idempotency_key_exists", False)


def idempotency_key_response(request):
    return getattr(request, "idempotency_key_response", None)


def get_idempotency_key_settings():
    return getattr(settings, "IDEMPOTENCY_KEY", dict())


def get_encoder_class():
    return module_loading.import_string(
        get_idempotency_key_settings().get(
            "ENCODER_CLASS", "idempotency_key.encoders.BasicKeyEncoder"
        )
    )


def get_conflict_code():
    return get_idempotency_key_settings().get(
        "CONFLICT_STATUS_CODE", status.HTTP_409_CONFLICT
    )


def get_storage_settings():
    return get_idempotency_key_settings().get("STORAGE", dict())


def get_storage_class():
    return module_loading.import_string(
        get_storage_settings().get("CLASS", "idempotency_key.storage.MemoryKeyStorage")
    )


def get_storage_cache_name():
    return get_storage_settings().get("CACHE_NAME", "default")


def get_storage_store_on_statuses():
    return get_storage_settings().get(
        "STORE_ON_STATUSES",
        [
            status.HTTP_200_OK,
            status.HTTP_201_CREATED,
            status.HTTP_202_ACCEPTED,
            status.HTTP_203_NON_AUTHORITATIVE_INFORMATION,
            status.HTTP_204_NO_CONTENT,
            status.HTTP_205_RESET_CONTENT,
            status.HTTP_206_PARTIAL_CONTENT,
            status.HTTP_207_MULTI_STATUS,
        ],
    )


def get_lock_settings():
    return get_idempotency_key_settings().get("LOCK", dict())


def get_lock_class():
    return module_loading.import_string(
        get_lock_settings().get("CLASS", "idempotency_key.locks.basic.ThreadLock")
    )


def get_lock_location():
    return get_lock_settings().get("LOCATION", "Redis://localhost:6379/1")


def get_lock_timeout():
    return get_lock_settings().get("TIMEOUT", 0.1)  # default to 100ms


def get_lock_enable():
    return get_lock_settings().get("ENABLE", True)


def get_lock_time_to_live():
    return get_lock_settings().get("TTL", 300)  # default to 5 minutes


def get_lock_name():
    return get_lock_settings().get("NAME", "MyLock")


def get_header_name():
    return get_idempotency_key_settings().get("HEADER", "HTTP_IDEMPOTENCY_KEY")



from functools import wraps

from idempotency_key import utils


# NOTE:
# The following decorators must be specified BEFORE the @api_view decorator or the function will not be marked
# correctly.
#
# i.e:
#
# @idempotency_key
# @api_view(['POST'])
# def my_view_func()
#   ...


def idempotency_key(*args, optional=False, cache_name=None):
    """
    Allows an optional cache name to be specified so that different cache settings can be used on a per-view function
    basis.
    :param args: optional arguments. This can contain the view function object if cache_name is not specified
    :param optional: Mark idempotency key header as optional
    :param cache_name: The name of the cache to use from the settings file under CACHES={...}
    :return: wrapped function
    """

    def _idempotency_key(view_func):
        """
        Mark a view function as requiring idempotency key protection but the view should control the response.
        """

        @wraps(view_func)
        def wrapped_view(*args, **kwargs):
            return view_func(*args, **kwargs)

        wrapped_view.idempotency_key = True
        wrapped_view.idempotency_key_optional = optional

        if cache_name:
            wrapped_view.idempotency_key_cache_name = cache_name
            utils.get_storage_class().validate_storage(cache_name)

        return wrapped_view

    # if there is an argument passed and it is a callable then this will be the view function object so pass it to
    # the wrapper
    if len(args) > 0 and callable(args[0]):
        return _idempotency_key(args[0])

    # otherwise just return the wrapper object
    return _idempotency_key


def idempotency_key_exempt(view_func):
    """
    Mark a view function as being exempt from the idempotency key protection.
    """

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.idempotency_key_exempt = True
    return wraps(view_func)(wrapped_view)


def idempotency_key_manual(view_func):
    """
    Mark a view function as requiring idempotency key protection but the view should control the response.
    """

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.idempotency_key_manual = True
    return wraps(view_func)(wrapped_view)



import logging

from django.core.exceptions import ImproperlyConfigured

from idempotency_key import status
from idempotency_key import utils
from idempotency_key.exceptions import (
    DecoratorsMutuallyExclusiveError,
    bad_request,
    resource_locked,
)

logger = logging.getLogger("django-idempotency-key.idempotency_key.middleware")


class IdempotencyKeyMiddleware:
    """
    This middleware class assumes that all non-safe HTTP methods will require an
    idempotency key to be specified in the header.
    View functions can opt-out using the @idempotency_key_exempt decorator
    """

    def __init__(self, get_response=None):
        self.get_response = get_response
        self.storage = utils.get_storage_class()()
        self.encoder = utils.get_encoder_class()()
        self.storage_lock = utils.get_lock_class()()

    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        response = self.process_response(request, response)
        return response

    @staticmethod
    def _reject(request, reason):
        response = bad_request(request, None)
        logger.debug(
            "Bad Request (%s): %s",
            reason,
            request.path,
            extra={"status_code": 400, "request": request},
        )
        return response

    def _set_flags_from_callback(self, request, callback):
        # If there is an actions attribute then the function is wrapped in a DRF viewset
        func_name = callback.__name__
        if hasattr(callback, "actions"):
            func_name = callback.actions[request.method.lower()]
            # get a reference to the function to access any attributes we might be
            # interested in.
            callback = getattr(callback.cls, func_name, callback)

        idempotency_key = getattr(callback, "idempotency_key", None)
        idempotency_key_optional = getattr(callback, "idempotency_key_optional", False)
        idempotency_key_exempt = getattr(callback, "idempotency_key_exempt", False)
        idempotency_key_manual = getattr(callback, "idempotency_key_manual", False)
        idempotency_key_cache_name = getattr(
            callback, "idempotency_key_cache_name", utils.get_storage_cache_name()
        )

        if idempotency_key and idempotency_key_exempt:
            raise DecoratorsMutuallyExclusiveError(
                "@idempotency_key and @idempotency_key_exempt decorators are mutually "
                'exclusive for function "{}"'.format(func_name)
            )

        if idempotency_key_manual and idempotency_key_exempt:
            raise DecoratorsMutuallyExclusiveError(
                "@idempotency_key_manual and @idempotency_key_exempt decorators are "
                'mutually exclusive for function "{}"'.format(func_name)
            )

        request.idempotency_key_optional = idempotency_key_optional
        request.idempotency_key_exempt = idempotency_key_exempt
        request.idempotency_key_manual = idempotency_key_manual
        request.idempotency_key_cache_name = idempotency_key_cache_name

    def perform_generate_response(self, request, encoded_key):
        # Check if a response already exists for the encoded key
        key_exists, response = self.storage.retrieve_data(
            request.idempotency_key_cache_name, encoded_key
        )

        # add the key exists result and the original request if it exists
        request.idempotency_key_exists = key_exists
        request.idempotency_key_response = response

        # If not manual override and the key already exists
        if not request.idempotency_key_manual and key_exists:
            # Get the required return status code from settings
            status_code = utils.get_conflict_code()
            # if None then return whatever the status code was originally otherwise use
            # the specified status code
            if status_code is not None:
                response.status_code = status_code
            return response

        return None

    def generate_response(self, request, encoded_key, lock=None):
        if lock is None:
            lock = utils.get_lock_enable()

        if not lock:
            return self.perform_generate_response(request, encoded_key)

        # If there was a timeout for a lock on the storage object then return a
        # HTTP_423_LOCKED
        if not self.storage_lock.acquire():
            return resource_locked(request, None)

        try:
            return self.perform_generate_response(request, encoded_key)
        finally:
            self.storage_lock.release()

    def process_request(self, request):
        key = request.META.get(utils.get_header_name())
        if key is not None:
            request.META["IDEMPOTENCY_KEY"] = key

        # Use this attribute to check that process_view has been called.
        request.idempotency_key_done = False

    def process_view(self, request, callback, _callback_args, _callback_kwargs):
        self._set_flags_from_callback(request, callback)

        # signal the process_view has been called
        request.idempotency_key_done = True

        # Assume that anything defined as 'safe' by RFC7231 is exempt or if exempt is
        # specified directly
        if request.idempotency_key_exempt or request.method in (
            "GET",
            "HEAD",
            "OPTIONS",
            "TRACE",
        ):
            request.idempotency_key_exempt = True
            return None

        # At this point the view function is not exempt so mark it as such
        request.idempotency_key_exempt = False

        key = request.META.get("IDEMPOTENCY_KEY")
        if key is None:
            if request.idempotency_key_optional:
                request.idempotency_key_exempt = True
                return None
            return self._reject(
                request,
                "Idempotency key is required and was not specified in the header.",
            )

        # encode the key and add it to the request
        encoded_key = request.idempotency_key_encoded_key = self.encoder.encode_key(
            request, key
        )

        # Generate the response
        return self.generate_response(request, encoded_key)

    def process_response(self, request, response):
        # If the response is not in the 20X range then return the response because at
        # this point protecting it with an idempotency key is meaningless.
        if response and response.status_code not in [
            status.HTTP_200_OK,
            status.HTTP_201_CREATED,
            status.HTTP_202_ACCEPTED,
            status.HTTP_203_NON_AUTHORITATIVE_INFORMATION,
            status.HTTP_204_NO_CONTENT,
            status.HTTP_205_RESET_CONTENT,
            status.HTTP_206_PARTIAL_CONTENT,
            status.HTTP_207_MULTI_STATUS,
        ]:
            return response

        # Make sure that process_view is called otherwise the use of idempotency keys
        # will be overridden without us knowing about it.
        if not getattr(request, "idempotency_key_done", False):
            raise ImproperlyConfigured(
                "Idempotency key middleware's 'process_view' function was not called! "
                "There maybe another middleware stopping this from happening which "
                "means your functions will not be properly protected with idempotency "
                "keys."
            )

        if getattr(request, "idempotency_key_exempt", True):
            return response

        if request.method not in ("GET", "HEAD", "OPTIONS", "TRACE"):
            # If the response matches that given by the store_on_statuses function then
            # store the data
            if response.status_code in utils.get_storage_store_on_statuses():
                self.storage.store_data(
                    request.idempotency_key_cache_name,
                    request.idempotency_key_encoded_key,
                    response,
                )

        return response


class ExemptIdempotencyKeyMiddleware(IdempotencyKeyMiddleware):
    """
    This middleware class assume all requests are exempt and do not require an
    idempotency key to be specified.
    View functions opt-in using the @idempotency_key or @idempotency_key_manual
    decorators.
    """

    def _set_flags_from_callback(self, request, callback):
        func_name = callback.__name__
        # If there is an actions attribute then the function is wrapped in a DRF viewset
        if hasattr(callback, "actions"):
            actual_func_name = callback.actions.get(request.method.lower())
            # if for some reason the method is not available in the viewset then just
            # proceed as normal and let the framework handle the problem.
            if actual_func_name is not None:
                func_name = actual_func_name

                # get a reference to the function to access any attributes we might be
                # interested in.
                callback = getattr(callback.cls, func_name, callback)

        idempotency_key = getattr(callback, "idempotency_key", False)
        idempotency_key_optional = getattr(callback, "idempotency_key_optional", False)
        idempotency_key_exempt = getattr(callback, "idempotency_key_exempt", None)
        idempotency_key_manual = getattr(callback, "idempotency_key_manual", False)
        idempotency_key_cache_name = getattr(
            callback, "idempotency_key_cache_name", utils.get_storage_cache_name()
        )

        if idempotency_key and idempotency_key_exempt:
            raise DecoratorsMutuallyExclusiveError(
                "@idempotency_key and @idempotency_key_exempt decorators are mutually "
                'exclusive for function "{}"'.format(func_name)
            )

        if idempotency_key_manual and idempotency_key_exempt:
            raise DecoratorsMutuallyExclusiveError(
                "@idempotency_key_manual and @idempotency_key_exempt decorators are "
                'mutually exclusive for function "{}"'.format(func_name)
            )

        request.idempotency_key_optional = idempotency_key_optional
        request.idempotency_key_exempt = idempotency_key_exempt or (
            idempotency_key_exempt is None
            and not idempotency_key_manual
            and not idempotency_key
        )
        request.idempotency_key_manual = idempotency_key_manual
        request.idempotency_key_cache_name = idempotency_key_cache_name



from django.http import JsonResponse

from idempotency_key import status


class MissingIdempotencyKeyError(Exception):
    def __init__(self, msg=None):
        if msg is None:
            msg = "Idempotency key cannot be None."
        super().__init__(msg)

    """
    Raised when an idempotency key has not been specified
    """
    pass


class DecoratorsMutuallyExclusiveError(Exception):
    pass


def bad_request(request, exception, *args, **kwargs):
    """
    Generic 400 error handler.
    """
    data = {"error": "Bad Request (400)"}
    return JsonResponse(data, status=status.HTTP_400_BAD_REQUEST)


def resource_locked(request, exception, *args, **kwargs):
    """
    Generic 423 error handler.
    """
    data = {"error": "Resource Locked (423)"}
    return JsonResponse(data, status=status.HTTP_423_LOCKED)



from redis import Redis

from idempotency_key import utils
from idempotency_key.locks.basic import IdempotencyKeyLock


class MultiProcessRedisLock(IdempotencyKeyLock):
    """
    Should be used if a lock is required across processes. Note that this class uses
    Redis in order to perform the lock.
    """

    def __init__(self):
        location = utils.get_lock_location()
        if location is None or location == "":
            raise ValueError("Redis server location must be set in the settings file.")

        self.redis_obj = Redis.from_url(location)
        self.storage_lock = self.redis_obj.lock(
            name=utils.get_lock_name(),
            # Time before lock is forcefully released.
            timeout=utils.get_lock_time_to_live(),
            blocking_timeout=utils.get_lock_timeout(),
        )

    def acquire(self, *args, **kwargs) -> bool:
        return self.storage_lock.acquire()

    def release(self):
        self.storage_lock.release()



import abc
import hashlib

from idempotency_key.exceptions import MissingIdempotencyKeyError


class IdempotencyKeyEncoder(object):
    @abc.abstractmethod
    def encode_key(self, request, key):
        raise NotImplementedError


class BasicKeyEncoder(IdempotencyKeyEncoder):
    def encode_key(self, request, key):
        if key is None:
            raise MissingIdempotencyKeyError()
        # Basic method for generating an encoded key
        m = hashlib.sha256()
        m.update(key.encode("UTF-8"))
        m.update(request.path_info.encode("UTF-8"))
        m.update(request.method.encode("UTF-8"))
        m.update(request.body)
        return m.hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

setup(
    name="django-idempotency-key",
    version="1.1.1",
    author="Del Hyman-Jones",
    author_email="dev@yoyowallet.com",
    description=(
        "Django middleware for idempotency key support in view and viewset functions."
    ),
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    platforms=["Any"],
    keywords=["django", "idempotency", "key", "idempotency-key", "middleware"],
    url="https://github.com/yoyowallet/django-idempotency-key",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Programming Language :: Python :: 3.6",
    ],
    install_requires=["Django>=2.0,<3"],
    project_urls={
        "Documentation": (
            "https://github.com/yoyowallet/django-idempotency-key/blob/master/README.md"
        ),
        "Source": "https://github.com/yoyowallet/django-idempotency-key",
    },
    packages=find_packages(exclude=["tests*"]),
    setup_requires=["setuptools>=38.6.0"],
    include_package_data=True,
    zip_safe=False,
)
